"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _keys = require("babel-runtime/core-js/object/keys");

var _keys2 = _interopRequireDefault(_keys);

var _asyncToGenerator2 = require("babel-runtime/helpers/asyncToGenerator");

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _ducky = require("ducky");

var _ducky2 = _interopRequireDefault(_ducky);

var _bluebird = require("bluebird");

var _bluebird2 = _interopRequireDefault(_bluebird);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*  the API mixin class  */
/*
**  GraphQL-Tools-Subscribe -- Subscription Framework for GraphQL-Tools
**  Copyright (c) 2016-2018 Ralf S. Engelschall <rse@engelschall.com>
**
**  Permission is hereby granted, free of charge, to any person obtaining
**  a copy of this software and associated documentation files (the
**  "Software"), to deal in the Software without restriction, including
**  without limitation the rights to use, copy, modify, merge, publish,
**  distribute, sublicense, and/or sell copies of the Software, and to
**  permit persons to whom the Software is furnished to do so, subject to
**  the following conditions:
**
**  The above copyright notice and this permission notice shall be included
**  in all copies or substantial portions of the Software.
**
**  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
**  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
**  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
**  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
**  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
**  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
**  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*  external dependencies  */
class gtsEvaluation {
    /*  stringify a scope record  */
    __recordStringify(rec) {
        let str = "";
        if (rec.srcType !== null && rec.srcId !== null && rec.srcAttr !== null) str += `${rec.srcType}#${rec.srcId}.${rec.srcAttr}->`;
        str += `${rec.op}(${rec.arity})->`;
        str += `${rec.dstType}#{${rec.dstIds.join(",")}}.{${rec.dstAttrs.join(",")}}`;
        return str;
    }

    /*  unstringify a scope record  */
    __recordUnstringify(str) {
        let m = str.match(/^(?:(.+?)#(.+?)\.(.+?)->)?(.+?)\((.+?)\)->(.+?)#\{(.*?)\}\.\{(.+?)\}$/);
        if (m === null) throw new Error(`invalid record string "${str}" (failed to parse)`);
        return {
            srcType: m[1] || null,
            srcId: m[2] || null,
            srcAttr: m[3] || null,
            op: m[4],
            arity: m[5],
            dstType: m[6],
            dstIds: m[7].split(","),
            dstAttrs: m[8].split(",")
        };
    }

    /*  serialize a list of records  */
    __recordsSerialize(records) {
        return records.map(record => {
            return this.__recordStringify(record);
        });
    }

    /*  unserialize a list of records  */
    __recordsUnserialize(records) {
        let result = null;
        if (_ducky2.default.validate(records, "[ string* ]")) {
            try {
                result = records.map(record => {
                    return this.__recordUnstringify(record);
                });
            } catch (ex) {
                /* ignore */
                result = null;
            }
        }
        return result;
    }

    /*  does a new scope (with mutation/write) outdate an old scope (with query/read only)  */
    __scopeOutdated(recordsNew, recordsOld) {
        /*  check if two lists overlap  */
        const overlap = (list1, list2) => {
            if (list1.length === 0 || list2.length === 0) return false;
            if (list1.length === 1 && list1[0] === "*") return true;
            if (list2.length === 1 && list2[0] === "*") return true;
            let index = {};
            list1.forEach(x => {
                index[x] = true;
            });
            let overlap = list2.filter(x => index[x]);
            return overlap.length > 0;
        };

        /*  iterate over all new and old records...  */
        for (let i = 0; i < recordsNew.length; i++) {
            let recNew = recordsNew[i];
            for (let j = 0; j < recordsOld.length; j++) {
                let recOld = recordsOld[j];
                let outdated = false;

                /*
                 *  CASE 1: modified entity (of arbitrary direct access)
                 *  old/query:    read(*)->Item#{1}.{id,name}
                 *  new/mutation: update/delete(*)->Item#{1}.{name}
                 */
                if ((recNew.op === "update" || recNew.op === "delete") && recOld.dstType === recNew.dstType && overlap(recOld.dstIds, recNew.dstIds) && overlap(recOld.dstAttrs, recNew.dstAttrs)) outdated = true;

                /*
                 *  CASE 2: modified entity list (of relationship traversal)
                 *  old/query     Card#1.items->read(*)->Item#{2}.{id,name}
                 *  new/mutation: update(*)->Card#{1}.{items}
                 */
                else if (recNew.op === "update" && recOld.srcType !== null && recOld.srcType === recNew.dstType && overlap([recOld.srcId], recNew.dstIds) && overlap([recOld.srcAttr], recNew.dstAttrs)) outdated = true;

                    /*
                     *  CASE 3: modified entity list (of direct query)
                     *  old/query     read(many/all)->Item#{*}.{id,name}
                     *  new/mutation: create/update/delete(*)->Item#{*}.{name}
                     */
                    else if ((recNew.op === "create" || recNew.op === "delete") && recOld.srcType === null && (recOld.arity === "many" || recOld.arity === "all") && recOld.dstType === recNew.dstType && overlap(recOld.dstAttrs, recNew.dstAttrs)) outdated = true;

                /*  report outdate combination  */
                if (outdated) {
                    this.emit("scope-outdated", { old: recordsOld[j], new: recordsNew[i] });
                    let recOld = this.__recordStringify(recordsOld[j]);
                    let recNew = this.__recordStringify(recordsNew[i]);
                    this.emit("debug", `scope-outdated old=${recOld} new=${recNew}`);
                    return true;
                }
            }
        }

        /*  ...else the scope is still valid (not outdated)  */
        return false;
    }

    /*  process a committed scope  */
    __scopeProcess(scope) {
        var _this = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  determine parameters  */
            let sid = scope.sid;
            let cid = scope.connection !== null ? scope.connection.cid : `${_this.uuid}:none`;

            /*  filter out write records in the scope  */
            let recordsWrite = scope.records.filter(function (record) {
                return record.op.match(/^(?:create|update|delete)$/);
            });

            /*  queries (scopes without writes)...  */
            if (recordsWrite.length === 0) {
                if (scope.state === "subscribed") {
                    /*  ...with subscriptions are remembered  */
                    let rec = yield _this.keyval.get(`sid:${sid},rec`);
                    let recNew = _this.__recordsSerialize(scope.records);
                    if (rec === undefined || rec !== recNew) yield _this.keyval.put(`sid:${sid},rec`, recNew);
                    let pid = yield _this.keyval.get(`sid:${sid},cid:${cid}`);
                    if (pid === undefined || pid !== process.pid) yield _this.keyval.put(`sid:${sid},cid:${cid}`, process.pid);
                    _this.emit("debug", `scope-store-update sid=${sid} cid=${cid}`);
                } else {
                    /*  ...without subscriptions can be just destroyed
                        (and the processing short-circuited)  */
                    scope.destroy();
                }
            }

            /*  mutations (scopes with writes) might outdate queries (scopes with reads)  */
            else {
                    /*  iterate over all stored scope records  */
                    let keys = yield _this.keyval.keys("sid:*,rec");
                    let sids = keys.map(function (key) {
                        return key.replace(/^sid:(.+?),rec$/, "$1");
                    });
                    let outdatedSids = [];
                    yield _bluebird2.default.each(sids, (() => {
                        var _ref = (0, _asyncToGenerator3.default)(function* (otherSid) {
                            let records = yield _this.keyval.get(`sid:${otherSid},rec`);
                            if (records !== undefined) {
                                records = _this.__recordsUnserialize(records);
                                if (records === null) yield _this.keyval.del(`sid:${otherSid},rec`);else if (_this.__scopeOutdated(recordsWrite, records)) outdatedSids.push(otherSid);
                            }
                        });

                        return function (_x) {
                            return _ref.apply(this, arguments);
                        };
                    })());

                    /*  externally publish ids of outdated queries to all instances
                        (comes in on all instances via __scopeOutdatedEvent below)  */
                    if (outdatedSids.length > 0) {
                        _this.emit("debug", `scope-outdated-send sids=${outdatedSids.join(",")}`);
                        _this.pubsub.publish("outdated", outdatedSids);
                    }
                }
        })();
    }

    /*  process an outdated event  */
    __scopeOutdatedEvent(sids) {
        this.emit("debug", `scope-outdated-receive sids=${sids.join(",")}`);
        this.connections.forEach(conn => {
            let outdated = {};
            conn.scopes.forEach(scope => {
                if (sids.indexOf(scope.sid) >= 0) {
                    if (scope.state === "subscribed") outdated[scope.sid] = true;else if (scope.state === "paused") scope.outdated = true;
                }
            });
            outdated = (0, _keys2.default)(outdated);
            if (outdated.length > 0) {
                this.emit("debug", `scope-outdated-notify sids=${outdated.join(",")}`);
                conn.notify(outdated);
            }
        });
    }

    /*  destroy a scope  */
    __scopeDestroy(scope) {
        var _this2 = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  determine parameters  */
            let sid = scope.sid;
            let cid = scope.connection !== null ? scope.connection.cid : `${_this2.uuid}:none`;

            /*  scope records with no more corresponding subscriptions are deleted  */
            yield _this2.keyval.del(`sid:${sid},cid:${cid}`);
            let keys = yield _this2.keyval.keys(`sid:${sid},cid:*`);
            if (keys.length === 0) yield _this2.keyval.del(`sid:${sid},rec`);
            _this2.emit("debug", `scope-store-delete sid=${sid} cid=${cid}`);
        })();
    }

    /*  dump current information  */
    dump() {
        var _this3 = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  determine information in store  */
            yield _this3.keyval.acquire();
            let info = { sids: {} };
            let keys = yield _this3.keyval.keys("sid:*,cid:*");
            keys.forEach(function (key) {
                let sid = key.replace(/^sid:(.+?),cid:.+?$/, "$1");
                let cid = key.replace(/^sid:.+?,cid:(.+?)$/, "$1");
                if (info.sids[sid] === undefined) info.sids[sid] = { cids: [], records: [] };
                info.sids[sid].cids.push(cid);
            });
            keys = yield _this3.keyval.keys("sid:*,rec");
            let sids = keys.map(function (key) {
                return key.replace(/^sid:(.+?),rec$/, "$1");
            });
            yield _bluebird2.default.each(sids, (() => {
                var _ref2 = (0, _asyncToGenerator3.default)(function* (sid) {
                    let records = yield _this3.keyval.get(`sid:${sid},rec`);
                    if (records !== undefined) {
                        records = _this3.__recordsUnserialize(records);
                        if (records === null) yield _this3.keyval.del(`sid:${sid},rec`);else info.sids[sid].records = info.sids[sid].records.concat(records);
                    }
                });

                return function (_x2) {
                    return _ref2.apply(this, arguments);
                };
            })());
            yield _this3.keyval.release();

            /*  dump information  */
            let dump = "";
            (0, _keys2.default)(info.sids).forEach(function (sid) {
                dump += `Scope { sid: ${sid} }\n`;
                info.sids[sid].cids.forEach(function (cid) {
                    dump += `    Connection { cid: ${cid} }\n`;
                });
                info.sids[sid].records.forEach(function (record) {
                    let rec = _this3.__recordStringify(record);
                    dump += `    Record { rec: ${rec} }\n`;
                });
            });
            return dump;
        })();
    }

    /*  flush all information  */
    flush() {
        var _this4 = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  flush all external storage information  */
            let keys = yield _this4.keyval.keys("sid:*,cid:*");
            yield _bluebird2.default.each(keys, (() => {
                var _ref3 = (0, _asyncToGenerator3.default)(function* (key) {
                    _this4.keyval.del(key);
                });

                return function (_x3) {
                    return _ref3.apply(this, arguments);
                };
            })());
            keys = yield _this4.keyval.keys("sid:*,rec");
            yield _bluebird2.default.each(keys, (() => {
                var _ref4 = (0, _asyncToGenerator3.default)(function* (key) {
                    _this4.keyval.del(key);
                });

                return function (_x4) {
                    return _ref4.apply(this, arguments);
                };
            })());
        })();
    }
}
exports.default = gtsEvaluation;
