"use strict";

var _asyncToGenerator2 = require("babel-runtime/helpers/asyncToGenerator");

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _assign = require("babel-runtime/core-js/object/assign");

var _assign2 = _interopRequireDefault(_assign);

var _es = require("aggregation/es6");

var _es2 = _interopRequireDefault(_es);

var _eventemitter = require("eventemitter3");

var _eventemitter2 = _interopRequireDefault(_eventemitter);

var _ipcPubsub = require("ipc-pubsub");

var _ipcPubsub2 = _interopRequireDefault(_ipcPubsub);

var _ipcKeyval = require("ipc-keyval");

var _ipcKeyval2 = _interopRequireDefault(_ipcKeyval);

var _pureUuid = require("pure-uuid");

var _pureUuid2 = _interopRequireDefault(_pureUuid);

var _gts1Tracking = require("./gts-1-tracking");

var _gts1Tracking2 = _interopRequireDefault(_gts1Tracking);

var _gts2Resolver = require("./gts-2-resolver");

var _gts2Resolver2 = _interopRequireDefault(_gts2Resolver);

var _gts3Evaluation = require("./gts-3-evaluation");

var _gts3Evaluation2 = _interopRequireDefault(_gts3Evaluation);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*  the API class  */
/*
**  GraphQL-Tools-Subscribe -- Subscription Framework for GraphQL-Tools
**  Copyright (c) 2016-2018 Ralf S. Engelschall <rse@engelschall.com>
**
**  Permission is hereby granted, free of charge, to any person obtaining
**  a copy of this software and associated documentation files (the
**  "Software"), to deal in the Software without restriction, including
**  without limitation the rights to use, copy, modify, merge, publish,
**  distribute, sublicense, and/or sell copies of the Software, and to
**  permit persons to whom the Software is furnished to do so, subject to
**  the following conditions:
**
**  The above copyright notice and this permission notice shall be included
**  in all copies or substantial portions of the Software.
**
**  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
**  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
**  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
**  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
**  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
**  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
**  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*  external dependencies  */
class GraphQLToolsSubscribe extends (0, _es2.default)(_gts1Tracking2.default, _gts2Resolver2.default, _gts3Evaluation2.default) {
    /*  the class constructor  */
    constructor(options = {}) {
        super();
        this.options = (0, _assign2.default)({
            pubsub: "spm",
            keyval: "spm"
        }, options);
        this.keyval = new _ipcKeyval2.default(this.options.keyval);
        this.pubsub = new _ipcPubsub2.default(this.options.pubsub);
        this.subscription = null;
        this.emitter = new _eventemitter2.default();
        this.uuid = new _pureUuid2.default(1).format();
    }

    /*  provide event listening  */
    on(name, handler) {
        this.emitter.on(name, handler);
        return () => {
            this.emitter.removeListener(name, handler);
        };
    }

    /*  provide event emitting  */
    emit(name, ...args) {
        this.emitter.emit(name, ...args);
        return this;
    }

    /*  open service  */
    open() {
        var _this = this;

        return (0, _asyncToGenerator3.default)(function* () {
            yield _this.keyval.open();
            yield _this.pubsub.open();
            _this.subscription = yield _this.pubsub.subscribe("outdated", function (sids) {
                _this.__scopeOutdatedEvent(sids);
            });
            return _this;
        })();
    }

    /*  close service  */
    close() {
        var _this2 = this;

        return (0, _asyncToGenerator3.default)(function* () {
            _this2.subscription.unsubscribe();
            _this2.subscription = null;
            yield _this2.keyval.close();
            yield _this2.pubsub.close();
            return _this2;
        })();
    }
}

/*  export the traditional way for interoperability reasons
    (as Babel would export an object with a 'default' field)  */


/*  internal dependencies  */
module.exports = GraphQLToolsSubscribe;
